#include "stdafx.h"
#include "TIM.h"
#include "Extract.h"
#include "global.h"

const unsigned int BufSize = 0x400000; // 4mb
unsigned int imagecount = 0;
unsigned int curoff = 0;
char* filename;
FILE* outlog = NULL;
unsigned int TimStart = 0;
unsigned int TimEnd = 0;

bool ExtractTIMs(const char* TargetFile, unsigned int Start, unsigned int End, unsigned int& ExtractedImages)
{
	ExtractedImages = 0;
	imagecount = 0;
	filename = (char*)TargetFile;
	curoff = 0;
	TimStart = Start;
	TimEnd = End;

	FILE* file = fopen(TargetFile, "rb");
	if(file == NULL)
		return false;

	char tbuf[255];
	sprintf(tbuf, "%s.log", TargetFile);
	outlog = fopen(tbuf, "w");

	unsigned int fsize;
	fseek(file, 0, SEEK_END);
	fsize = ftell(file);
	fseek(file, 0, SEEK_SET);

	unsigned bufs = fsize / BufSize + 1;
	unsigned char* buf = new unsigned char[BufSize];
	unsigned int sizeleft = fsize;
	unsigned int data = 0;

	for(unsigned int i = 0; i < bufs, imagecount <= TimEnd; i++)
	{
		if(sizeleft > BufSize)
		{
			fread(buf, 1, BufSize, file);
			sizeleft -= BufSize;
			data = BufSize;
		}
		else
		{
			fread(buf, 1, sizeleft, file);
			memset(buf+sizeleft, 0, BufSize - sizeleft);
			data = sizeleft;
			sizeleft = 0;
		}

		if(data <= 4)
			break;

		BYTE* d = buf;
		unsigned int imgsize;
		for(unsigned int j = 0; j < data - 4; j++, d++)
		{
			if(*(DWORD*)d == TIMFLAG)
			{
				curoff = i*BufSize+j;
				if((imgsize = ExtractTIM(file, curoff)) != -1)
				{
					j += imgsize-1;
					d += imgsize-1;
					if(imagecount > TimEnd)
						break;
				}
			}
		}

	}

	delete[] buf;
	fclose(file);
	fclose(outlog);

	if(imagecount > TimStart)
		ExtractedImages = imagecount - TimStart;
	else
		ExtractedImages = 0;

	return true;
}

unsigned int ExtractTIM(FILE* file, unsigned int offset)
{
	unsigned int prevoff = ftell(file);
	fseek(file, offset, SEEK_SET);

	unsigned int Id, BppId;
	fread(&Id, 4, 1, file);
	fread(&BppId, 4, 1, file);

	unsigned int size = 0;

	if(Id == TIMFLAG)
	{
		//if(BppId == _4BPPFLAG || BppId == _4BPPNCFLAG || BppId == _8BPPFLAG || BppId == _8BPPNCFLAG || BppId == _16BPPFLAG)
		if(BppId == _4BPPFLAG || BppId == _8BPPFLAG || BppId == _16BPPFLAG)
		{
			size = ReadTIM(file, BppId);
			fseek(file, prevoff, SEEK_SET);
			return size;
		}
	}

	fseek(file, prevoff, SEEK_SET);

	return -1;
}

unsigned int ReadTIM(FILE* file, unsigned int Type)
{
	TIM tim;
	memset(&tim, 0, sizeof(TIM));
	tim.ImageType = Type;

	// 4BPP / 8BPP headers
	if(Type == _4BPPFLAG || Type == _4BPPNCFLAG || Type == _8BPPFLAG || Type == _4BPPNCFLAG)
	{
		TIMCLUTHEADER1* hdr1 = new TIMCLUTHEADER1;
		tim.FirstHeader = (void*)hdr1;
		hdr1->BppId = Type;
		hdr1->TimId = TIMFLAG;

		if((sizeof(TIMCLUTHEADER1) - 8) != fread((char*)hdr1 + 8, 1, sizeof(TIMCLUTHEADER1) - 8, file))
		{
			//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
			DeleteTim(&tim);
			return -1;
		}
	}

	// 4BPP / 8BPP CLUT
	if(Type == _4BPPFLAG || Type == _8BPPFLAG) // Read CLUT
	{
		TIMCLUTHEADER1* hdr1 = (TIMCLUTHEADER1*)tim.FirstHeader;
		if(Type == _4BPPFLAG)
		{
			if(hdr1->ClutCount*hdr1->ClutColors*2 != hdr1->ClutSize - 12) // Verify
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}
           
			tim.Clut = new unsigned char[hdr1->ClutSize - 12];
			if(hdr1->ClutSize - 12 != fread(tim.Clut, 1, hdr1->ClutSize - 12, file))
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}
		}
		if(Type == _8BPPFLAG)
		{
			if(hdr1->ClutCount*hdr1->ClutColors*2 != hdr1->ClutSize - 12) // Verify
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}
           
			tim.Clut = new unsigned char[hdr1->ClutSize - 12];
			if(hdr1->ClutSize - 12 != fread(tim.Clut, 1, hdr1->ClutSize - 12, file))
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}
		}
	}

	// 4BPP / 8BPP second header and image data
	if(Type == _4BPPFLAG || Type == _4BPPNCFLAG || Type == _8BPPFLAG || Type == _4BPPNCFLAG)
	{
		TIMCLUTHEADER2* hdr2 = new TIMCLUTHEADER2;
		tim.SecondHeader = hdr2;
		if(sizeof(TIMCLUTHEADER2) != fread(hdr2, 1, sizeof(TIMCLUTHEADER2), file))
		{
			//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
			DeleteTim(&tim);
			return -1;
		}

		if(Type == _4BPPFLAG || Type == _4BPPNCFLAG)
		{
			// try to calc image size
			if(!(hdr2->ImageDataSize-12 == hdr2->ImageHeight * hdr2->ImageWidth * 4 || 
			   hdr2->ImageDataSize-12 == hdr2->ImageHeight * hdr2->ImageWidth * 2 ||
			   hdr2->ImageDataSize-12 == hdr2->ImageHeight * hdr2->ImageWidth))
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}

			tim.Image = new unsigned char[hdr2->ImageDataSize - 12];
			if(hdr2->ImageDataSize-12 != fread(tim.Image, 1, hdr2->ImageDataSize-12, file))
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}
		}
		else if(Type == _8BPPFLAG || Type == _8BPPNCFLAG)
		{
			// try to calc image size
			if(!(hdr2->ImageDataSize-12 == hdr2->ImageHeight * hdr2->ImageWidth * 2 || 
			   hdr2->ImageDataSize-12 == hdr2->ImageHeight * hdr2->ImageWidth * 4 ||
			   hdr2->ImageDataSize-12 == hdr2->ImageHeight * hdr2->ImageWidth))
			{
				//printf("%u %u\n", hdr2->ImageDataSize-12, hdr2->ImageHeight * hdr2->ImageWidth * 2);
				DeleteTim(&tim);
				return -1;
			}

			tim.Image = new unsigned char[hdr2->ImageDataSize - 12];
			if(hdr2->ImageDataSize-12 != fread(tim.Image, 1, hdr2->ImageDataSize-12, file))
			{
				//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
				DeleteTim(&tim);
				return -1;
			}
		}
	}

	// 16BPP
	if(Type == _16BPPFLAG)
	{
		TIMLINEARHEADER* hdr = new TIMLINEARHEADER;
		tim.FirstHeader = hdr;
		hdr->TimId = TIMFLAG;
		hdr->BppId = Type;
		if(sizeof(TIMLINEARHEADER)-8 != fread((char*)hdr+8, 1, sizeof(TIMLINEARHEADER)-8, file))
		{
			//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
			DeleteTim(&tim);
			return -1;
		}

		// try to calc image size
		if(!(hdr->ImageDataSize-12 == hdr->ImageHeight * hdr->ImageWidth * 2 || 
			  hdr->ImageDataSize-12 == hdr->ImageHeight * hdr->ImageWidth * 4 ||
			  hdr->ImageDataSize-12 == hdr->ImageHeight * hdr->ImageWidth))
		{
			//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
			DeleteTim(&tim);
			return -1;
		}

		tim.Image = new unsigned char[hdr->ImageDataSize-12];
		if(hdr->ImageDataSize-12 != fread(tim.Image, 1, hdr->ImageDataSize-12, file))
		{
			//printf("Failed verify/read on %s %u %d\n", __FILE__, __LINE__, curoff);
			DeleteTim(&tim);
			return -1;
		}
	}


	// Dump TIM
	unsigned int size = DumpTim(&tim);

	DeleteTim(&tim);
	return size;
}

unsigned int DumpTim(TIM* tim)
{
	if(imagecount < TimStart || imagecount > TimEnd) // out of range
	{
		imagecount++;
		return -1;
	}

	imagecount++;

	unsigned int size = 0;
	char buf[255];
	sprintf(buf, "%s%u.tim", filename, imagecount+1);
	FILE* out = fopen(buf, "wb");
	if(out == NULL)
		return -1;

	switch(tim->ImageType)
	{
	case _4BPPFLAG: case _8BPPFLAG:
		fwrite(tim->FirstHeader, 1, sizeof(TIMCLUTHEADER1), out);
		fwrite(tim->Clut, 1, ((TIMCLUTHEADER1*)tim->FirstHeader)->ClutSize-12, out);
		fwrite(tim->SecondHeader, 1, sizeof(TIMCLUTHEADER2), out);
		fwrite(tim->Image, 1, ((TIMCLUTHEADER2*)tim->SecondHeader)->ImageDataSize-12, out);
		fprintf(outlog, "%s %u %u\n", buf, curoff, ftell(out));
		break;

	case _4BPPNCFLAG: case _8BPPNCFLAG:
		fwrite(tim->FirstHeader, 1, sizeof(TIMCLUTHEADER1), out);
		fwrite(tim->SecondHeader, 1, sizeof(TIMCLUTHEADER2), out);
		fwrite(tim->Image, 1, ((TIMCLUTHEADER2*)tim->SecondHeader)->ImageDataSize-12, out);
		fprintf(outlog, "%s %u %u\n", buf, curoff, ftell(out));
		break;

	case _16BPPFLAG:
		fwrite(tim->FirstHeader, 1, sizeof(TIMLINEARHEADER), out);
		fwrite(tim->Image, 1, ((TIMLINEARHEADER*)tim->FirstHeader)->ImageDataSize-12, out);
		fprintf(outlog, "%s %u %u\n", buf, curoff, ftell(out));
		break;
	}

	size = ftell(out);
	fclose(out);
	return size;
}